Smalltalk organization addCategory: 'Collections-Large'!

IdentityDictionary subclass: #ExtendedAttributeDictionary
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Large'!

ExtendedAttributeDictionary class
	instanceVariableNames: ''!

ArrayedCollection variableSubclass: #LargeArray
	instanceVariableNames: 'size chunkSize '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Large'!

LargeArray class
	instanceVariableNames: ''!

IdentityDictionary subclass: #LargeIdentityDictionary
	instanceVariableNames: 'values '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Large'!

LargeIdentityDictionary class
	instanceVariableNames: ''!

ExtendedAttributeDictionary subclass: #TupleAttributeDictionary
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Large'!

TupleAttributeDictionary class
	instanceVariableNames: ''!

!ExtendedAttributeDictionary commentStamp: 'svp 8/13/2002 10:45' prior: 0!
I am an IdentityDictionary whose keys are held weakly.  I am designed for use when you need to potentially associate some attribute with any object in the system.  I am like a WeakIdentityKeyDictionary with the following differences:

- my hash function incorporates the object and the class' identity hash (therefore, objects that are subject to #become: operations should not be added as key's...or else you need to rehash after a become if the two object's classes are different) 
- I do not provide for finalization and rehash will eliminate all nil keys
- removeKey: performs much better than it does in WeakIdentityKeyDictionary (it does not require a full rehash of the dictionary)
- I aggressively nil out values of associations whose key is invalid (i.e. GC collected the key).  During any scan, or when removing a key, I will check for such associations and release the value
- #findElementOrNil: might answer a slot that contains an association with an invalid key if the element was not found.  Users of this method need to check for that.
- If I grow large, I use a LargeArray to store my associations.  This improves the performance of incremental GC, which degrades when there are very large objects in young space or when there are very large objects with references to objects in young space
- My hash is spread over 30 bits (and could thus map every single object in memory)...the hash function is about as efficient as possible given the limited number of identity hash bits

Use this class when you expect your dictionary to grow very large, when you don't need to finalize the weakly referenced objects, and/or when you need better key removal performance.  You may want to periodically #rehash instances to remove associations with nil keys.  Alternatively, you can #clearDeadValues if you do not have time to rehash the whole dictionary.!

!LargeArray commentStamp: 'svp 8/13/2002 13:49' prior: 0!
I am an array that is segmented into smaller arrays (called chunks). My chunks can in general be any indexable object.  My indexble slots actually contain the chunks, which can be accessed with basicAt: and basicAt:put:.  The methods at: and at:put: access my elements.

Here are a few examples:

	To create a large array for storing ordinary objects:
	LargeArray new: 1000 chunkSize: 10 arrayClass: Array

	To create a large byte array:
	LargeArray new: 1000 chunkSize: 10 arrayClass: ByteArray

	To create a large string:
	LargeArray new: 1000 chunkSize: 10 arrayClass: String

	To create an empty large array:
	LargeArray new

	To create a WriteStream on a LargeArray of Characters:
	WriteStream on: (LargeArray new: 0 chunkSize: 100 arrayClass: String)
!

!LargeIdentityDictionary commentStamp: 'svp 8/23/2002 20:14' prior: 0!
I am just like an IdentityDictionary except:

- I incorporate each key's class in the hash function (thus if you become: a key such that the two objects involved are not of the same class, you will need to rehash the dictionary (or remove and re-insert the affected key)

- I use two parallel LargeArrays with a chunkSize of 1024 to make large instances of me play nicely with Squeak's garbage collector.  One array (array) stores the keys, and the other (values) stores the values.  This results in a space savings of 2 words per association.  This also means that there are no Association instances to hold for repeated lookups of the same key.
!

!TupleAttributeDictionary commentStamp: 'svp 8/12/2002 23:39' prior: 0!
I am like my parent in that I can grow efficiently.  I am different in that my keys are tuples (one or more objects).  My keys are held weakly (if any item in a tuple is collected, then that association is no longer valid and lookups on that tuple will fail).  Lookup is based on the identity of each item in the tuple.

Examples:

	dict at: #(10 11) put: (Point x: 10 y: 11).
	dict at: {vendor. product} put: price.

I can be more efficient that performing multiple dictionary lookups to find data associated with more than one object. !

!ExtendedAttributeDictionary methodsFor: 'accessing' stamp: 'svp 8/12/2002 19:23'!
add: anAssociation
	self at: anAssociation key put: anAssociation value.
	^ anAssociation! !

!LargeIdentityDictionary methodsFor: 'accessing' stamp: 'svp 8/21/2002 15:04'!
add: anAssociation
	^ self at: anAssociation key put: anAssociation value! !

!LargeArray methodsFor: 'accessing' stamp: 'svp 8/12/2002 15:56'!
arrayClass

	^(self basicAt: 1) class! !

!ExtendedAttributeDictionary methodsFor: 'accessing' stamp: 'svp 8/13/2002 09:44'!
associationAt: key ifAbsent: aBlock 
	"Answer the association with the given key.
	If key is not found, return the result of evaluating aBlock."

	| index assoc |
	index _ self findElementOrNil: key.
	assoc _ array at: index.
	(assoc isNil or: [assoc key isNil]) ifTrue: [ ^ aBlock value ].
	^ assoc! !

!LargeIdentityDictionary methodsFor: 'accessing' stamp: 'svp 8/21/2002 15:10'!
associationAt: key ifAbsent: aBlock 
	"Answer the association with the given key.
	If key is not found, return the result of evaluating aBlock."

	| index k |
	index _ self findElementOrNil: key.
	k _ array at: index.
	k ifNil: [ ^ aBlock value ].
	^ Association key: k value: (values at: index)! !

!LargeArray methodsFor: 'accessing' stamp: 'svp 8/12/2002 20:26'!
at: index

	self pvtCheckIndex: index.
	^self noCheckAt: index.! !

!ExtendedAttributeDictionary methodsFor: 'accessing' stamp: 'svp 8/23/2002 17:47'!
at: key ifAbsent: aBlock 
	"Answer the value associated with the key or, if key isn't found,
	answer the result of evaluating aBlock."

	| assoc |
	assoc _ array at: (self findElementOrNil: key).
	(assoc isNil or: [self isKeyInvalid: assoc]) ifTrue: [^ aBlock value].
	^ assoc value
! !

!LargeIdentityDictionary methodsFor: 'accessing' stamp: 'svp 8/21/2002 15:09'!
at: key ifAbsent: aBlock 
	"Answer the value associated with the key or, if key isn't found,
	answer the result of evaluating aBlock."

	| k index |
	index := self findElementOrNil: key.
	k := array at: index.
	k ifNil: [^ aBlock value].
	^ values at: index! !

!ExtendedAttributeDictionary methodsFor: 'accessing' stamp: 'svp 8/13/2002 09:30'!
at: key put: anObject 
	"Set the value at key to be anObject.  If key is not found, create a new
	entry for key and set is value to anObject. Answer anObject."
	| index element |
	key isNil ifTrue:[^anObject].
	index _ self findElementOrNil: key.
	element _ array at: index.
	element == nil
		ifTrue: [self atNewIndex: index put: (WeakKeyAssociation key: key value: anObject)]
		ifFalse: 
			[element key ifNil: [element key: key]. 
			element value: anObject].
	^ anObject! !

!LargeArray methodsFor: 'accessing' stamp: 'svp 8/12/2002 20:27'!
at: index put: value
	
	self pvtCheckIndex: index.
	^self noCheckAt: index put: value! !

!LargeIdentityDictionary methodsFor: 'accessing' stamp: 'svp 8/21/2002 17:54'!
at: key put: anObject 
	"Set the value at key to be anObject.  If key is not found, create a
	new entry for key and set is value to anObject. Answer anObject."

	| index |
	index _ self findElementOrNil: key.
	(array at: index)
		ifNil: 
			[array at: index put: key.
			values at: index put: anObject.
			tally := tally + 1.
			self fullCheck]
		ifNotNil: 
			[values at: index put: anObject].
	^ anObject! !

!TupleAttributeDictionary methodsFor: 'accessing' stamp: 'svp 8/23/2002 17:14'!
at: key put: anObject 
	"Set the value at key to be anObject.  If key is not found, create a new
	entry for key and set is value to anObject. Answer anObject."
	| index element tuple |
	key isNil ifTrue:[^anObject].
	key isSequenceable
		ifTrue: [tuple := key]
		ifFalse: [tuple := Array with: key].
	index _ self findElementOrNil: tuple.
	element _ array at: index.
	element == nil
		ifTrue: 
			[self atNewIndex: index put: 
				(Association key: (WeakArray withAll: tuple) value: anObject)]
		ifFalse: 
			[(self isKeyValid: element) ifFalse: 
				[element key: (WeakArray withAll: tuple)].
			element value: anObject].
	^ anObject! !

!LargeArray methodsFor: 'replacement' stamp: 'svp 8/13/2002 13:40'!
basicReplaceFrom: start to: stop with: replacement startingAt: repStart 
	"This destructively replaces elements from start to stop in the receiver 
	starting at index, repStart, in the sequenceable collection, 
	replacementCollection. Answer the receiver. No range checks are 
	performed."
	
	"This implementation pushes the replacement down to the
	individual chunks.  This enables the replacement operation to be
	handled by primitives, thus getting the maximum possible performance."

	| startChunk startIndex stopChunk stopIndex cStart cStop cRepStart |
	startChunk := self pvtChunkIndexOf: start.
	startIndex := self pvtIndexOf: start inChunk: startChunk.
	stopChunk := self pvtChunkIndexOf: stop.
	stopIndex := self pvtIndexOf: stop inChunk: stopChunk.
	cStart := nil.

	startChunk to: stopChunk do: [ :chunkIndex |
		cRepStart := chunkIndex = startChunk 
					ifTrue: [repStart] 
					ifFalse: [cRepStart + (chunkSize - cStart + 1)].
		cStart := chunkIndex = startChunk ifTrue: [startIndex] ifFalse: [1].
		cStop := (chunkIndex = stopChunk)
					ifTrue: [stopIndex]
					ifFalse: [chunkSize].

		(self basicAt: chunkIndex) 
			replaceFrom: cStart to: cStop with: replacement startingAt: cRepStart].  
! !

!LargeArray class methodsFor: 'instance creation' stamp: 'svp 8/13/2002 13:18'!
bytesNew: size

	^self 
		new: size 
		chunkSize: ByteArray defaultLargeArrayChunkSize 
		arrayClass: ByteArray! !

!LargeArray class methodsFor: 'instance creation' stamp: 'svp 8/13/2002 13:18'!
bytesNew: size chunkSize: chunkSize

	^self 
		new: size 
		chunkSize: chunkSize 
		arrayClass: ByteArray! !

!ExtendedAttributeDictionary methodsFor: 'private' stamp: 'svp 12/6/2002 10:52'!
chunkSize
	"This value determines the maximum size of each chunk of our array"

	^128! !

!LargeArray methodsFor: 'accessing' stamp: 'svp 7/11/2002 16:08'!
chunkSize

	^chunkSize! !

!LargeIdentityDictionary methodsFor: 'private' stamp: 'svp 12/6/2002 10:52'!
chunkSize
	"This value determines the maximum size of the chunks in our array."

	^128! !

!ExtendedAttributeDictionary methodsFor: 'private' stamp: 'svp 8/23/2002 17:43'!
clearDeadValues
	"Nil out the value of any association whose key is nil (i.e. it's key was
	reclaimed by the garbage collector).  This is a faster alternative to #rehash, 
	which will remove the association itself."

	self associationsDo: [ :assn | (self isKeyInvalid: assn) ifTrue: [assn value: nil]]! !

!LargeIdentityDictionary methodsFor: 'private' stamp: 'svp 8/23/2002 16:18'!
compare: element with: anObject
	"Overridden because our elements are not associations...we are
	using two parallel arrays instead."

	^element == anObject! !

!TupleAttributeDictionary methodsFor: 'private' stamp: 'svp 8/23/2002 09:25'!
compare: element with: tuple

	element key size = tuple size ifFalse: [^false].
	1 to: tuple size do: [ :in | (element key at: in) == (tuple at: in) ifFalse: [^false]].
	^true! !

!LargeArray methodsFor: 'copying' stamp: 'svp 8/12/2002 11:17'!
copyInChunksOf: newChunkSize

	^(self species 
		new: self size 
		chunkSize: newChunkSize 
		arrayClass: self arrayClass)
			replaceFrom: 1
			to: size
			with: self
			startingAt: 1! !

!ExtendedAttributeDictionary methodsFor: 'private' stamp: 'svp 8/13/2002 10:05'!
deadKey

	^'Dead Key'! !

!LargeArray class methodsFor: 'accessing' stamp: 'svp 7/11/2002 16:34'!
defaultChunkSize

	^100! !

!LargeArray class methodsFor: 'accessing' stamp: 'svp 8/1/2002 15:09'!
defaultChunkSizeForFiles

	^8000! !

!ExtendedAttributeDictionary methodsFor: 'accessing' stamp: 'svp 8/13/2002 11:16'!
fasterKeys
	"This was taking some time in publishing and we didn't really need a Set"
	| answer |
	answer _ OrderedCollection new: self size.
	self keysDo: [:key | answer add: key].
	^ answer! !

!LargeIdentityDictionary methodsFor: 'accessing' stamp: 'svp 8/21/2002 14:19'!
fasterKeys
	"This was taking some time in publishing and we didn't really need a Set"
	| answer |
	answer _ OrderedCollection new: self size.
	self keysDo: [:key | answer add: key].
	^ answer! !

!ExtendedAttributeDictionary methodsFor: 'private' stamp: 'svp 8/13/2002 10:10'!
fixCollisionsFrom: index
	"The element at index has been removed and replaced by nil.
	This method moves forward from there, relocating any entries
	that had been placed below due to collisions with this one"

	| length oldIndex newIndex element deadKey |

	deadKey := self deadKey.
	oldIndex := index.
	length := array size.
	[oldIndex = length
			ifTrue: [oldIndex := 1]
			ifFalse: [oldIndex := oldIndex + 1].
	(element := self keyAt: oldIndex) == nil]
		whileFalse: 
			[element == deadKey 
				ifTrue: 
					["Take this opportunity to release the values of any GCed keys"
					(array at: oldIndex) value: nil.
					tally := tally - 1]
				ifFalse: 
					[newIndex _ self findElementOrNil: element.
					oldIndex = newIndex ifFalse: [self swap: oldIndex with: newIndex]]]! !

!LargeArray class methodsFor: 'instance creation' stamp: 'svp 8/1/2002 15:06'!
fromFileNamed: fileName

	^self fromFileStream: (FileStream oldFileNamed: fileName)
! !

!LargeArray class methodsFor: 'instance creation' stamp: 'svp 8/2/2002 16:11'!
fromFileStream: file

	| tmp chunkSize |
	chunkSize := self defaultChunkSizeForFiles.
	tmp := WriteStream on: 
		(self new: file size chunkSize: chunkSize arrayClass: String).
	[file atEnd] whileFalse:
		[tmp nextPutAll: (file next: chunkSize)].
	^tmp contents
	! !

!ExtendedAttributeDictionary methodsFor: 'private' stamp: 'svp 8/23/2002 17:39'!
getHashFor: object

	^array size > 4096 
		ifTrue: [object identityHash hashMultiply]
		ifFalse: [object identityHash]! !

!LargeIdentityDictionary methodsFor: 'private' stamp: 'svp 8/23/2002 16:15'!
getHashFor: anObject

	^(anObject identityHash * anObject class identityHash) hashMultiply.
! !

!TupleAttributeDictionary methodsFor: 'private' stamp: 'svp 8/23/2002 17:02'!
getHashFor: tuple

	^(1 to: tuple size) inject: 0 into: [ :hash :in |
		hash bitXor: (tuple at: in) identityHash hashMultiply].
! !

!ExtendedAttributeDictionary methodsFor: 'private' stamp: 'svp 8/23/2002 17:38'!
grow
	"Grow the elements array and reinsert the old elements"
	| oldElements newSize |
	oldElements _ array.
	newSize := array size + self growSize.
	array := LargeArray new: newSize chunkSize: self chunkSize arrayClass: Array.
	tally _ 0.
	oldElements do:
		[:each | each == nil ifFalse: [self noCheckAdd: each]]! !

!LargeIdentityDictionary methodsFor: 'accessing' stamp: 'svp 8/23/2002 17:36'!
grow
	"Grow the elements array and reinsert the old elements"
	| newSize chunkSize oldKeys oldValues |
	oldKeys := array.
	oldValues := values.
	newSize := array size + self growSize.
	chunkSize := self chunkSize.
	array := LargeArray new: newSize chunkSize: chunkSize arrayClass: Array.
	values := LargeArray new: newSize chunkSize: chunkSize arrayClass: Array.
	tally _ 0.
	oldKeys with: oldValues do:
		[:k :v | k == nil ifFalse: [self noCheckAt: k put: v]]! !

!ExtendedAttributeDictionary methodsFor: 'private' stamp: 'svp 8/12/2002 17:08'!
growthThreshold
	"When this dictionary has more associations than this, we will
	begin using LargeArrays for our array (with a chunkSize equal to
	this value)."

	^1024! !

!ExtendedAttributeDictionary methodsFor: 'initialization' stamp: 'svp 8/23/2002 17:36'!
init: n
	"Initialize array to an array size of n"

	array := LargeArray new: n chunkSize: self chunkSize arrayClass: Array.
	tally _ 0! !

!LargeIdentityDictionary methodsFor: 'accessing' stamp: 'svp 8/23/2002 17:37'!
init: n
	"Initialize array to an array size of n"

	| chunkSize |
	chunkSize := self chunkSize.
	array := LargeArray new: n chunkSize: chunkSize arrayClass: Array.
	values := LargeArray new: n chunkSize: chunkSize arrayClass: Array.
	tally _ 0! !

!LargeArray methodsFor: 'initialization' stamp: 'svp 8/13/2002 13:45'!
initChunkSize: aChunkSize size: aSize arrayClass: aClass

	| lastChunkSize |
	chunkSize := aChunkSize.
	size := aSize.
	1 to: (self basicSize - 1) do: [ :in | self basicAt: in put: (aClass new: chunkSize) ].
	lastChunkSize := size \\ chunkSize.
	lastChunkSize = 0 ifTrue: [lastChunkSize := chunkSize].
	size = 0 
		ifTrue: [self basicAt: 1 put: (aClass new: 0)]
		ifFalse: [self basicAt: self basicSize put: (aClass new: lastChunkSize)].! !

!ExtendedAttributeDictionary methodsFor: 'user interface' stamp: 'svp 8/22/2002 23:57'!
inspect
	"Because we are probably very big, don't attempt to use the dictionary
	inspector"

	self size < 100 ifTrue: [^super inspect].
	self basicInspect! !

!LargeIdentityDictionary methodsFor: 'user interface' stamp: 'svp 8/22/2002 23:51'!
inspect
	"Because we are probably very big, don't attempt to use the dictionary
	inspector"

	self size < 100 ifTrue: [^super inspect].
	self basicInspect! !

!ExtendedAttributeDictionary methodsFor: 'private' stamp: 'svp 8/23/2002 17:40'!
isKeyInvalid: assn

	^assn key == nil! !

!TupleAttributeDictionary methodsFor: 'private' stamp: 'svp 8/23/2002 17:11'!
isKeyInvalid: assn

	^assn key anySatisfy: [ :ea | ea == nil ]! !

!ExtendedAttributeDictionary methodsFor: 'private' stamp: 'svp 8/23/2002 17:40'!
isKeyValid: assn

	^(self isKeyInvalid: assn) not! !

!ExtendedAttributeDictionary methodsFor: 'private' stamp: 'svp 8/23/2002 17:45'!
keyAt: index
	"May be overridden by subclasses so that fixCollisions will work"
	| assn |
	^(assn _ array at: index)
		ifNil: [nil]
		ifNotNil: [(self isKeyValid: assn) 
					ifTrue: [assn key]
					ifFalse: [self deadKey]]! !

!TupleAttributeDictionary methodsFor: 'private' stamp: 'svp 8/13/2002 11:09'!
keyAtIdentityValue: value ifAbsent: exceptionBlock
	"Answer the key that is the external name for the argument, value. If 
	there is none, answer the result of evaluating exceptionBlock.
	Note: There can be multiple keys with the same value. Only one is returned."
 
	self associationsDo: 
		[:association | value == association value 
			ifTrue: [^Array withAll: association key]].
	^exceptionBlock value! !

!TupleAttributeDictionary methodsFor: 'private' stamp: 'svp 8/13/2002 11:09'!
keyAtValue: value ifAbsent: exceptionBlock
	"Answer the key that is the external name for the argument, value. If 
	there is none, answer the result of evaluating exceptionBlock."
 
	self associationsDo: 
		[:association | value = association value 
			ifTrue: [^ Array withAll: association key]].
	^ exceptionBlock value! !

!TupleAttributeDictionary methodsFor: 'private' stamp: 'svp 8/13/2002 11:10'!
keyForIdentity: anObject
	"If anObject is one of the values of the receive, return its key, else return nil.  Contrast #keyAtValue: in which there is only an equality check, here there is an identity check"

	^self keyAtIdentityValue: anObject ifAbsent: [nil]
! !

!ExtendedAttributeDictionary methodsFor: 'accessing' stamp: 'svp 8/13/2002 11:16'!
keys

	^self fasterKeys! !

!TupleAttributeDictionary methodsFor: 'private' stamp: 'svp 8/13/2002 11:11'!
keysAndValuesDo: aBlock
	^self associationsDo:[:assoc|
		aBlock value: (Array withAll: assoc key) value: assoc value].! !

!TupleAttributeDictionary methodsFor: 'private' stamp: 'svp 8/13/2002 11:13'!
keysAndValuesRemove: keyValueBlock
	"Removes all entries for which keyValueBlock returns true."
	"When removing many items, you must not do it while iterating over the dictionary, since it may be changing.  This method takes care of tallying the removals in a first pass, and then performing all the deletions afterward.  Many places in the sytem could be simplified by using this method."

	| removals key |
	removals _ OrderedCollection new.
	self associationsDo:
		[:assoc | (keyValueBlock value: (key := Array withAll: assoc key) value: assoc value)
			ifTrue: [removals add: key]].
 	removals do:
		[:aKey | self removeKey: aKey]! !

!ExtendedAttributeDictionary methodsFor: 'accessing' stamp: 'svp 8/23/2002 17:48'!
keysDo: aBlock 
	"Evaluate aBlock for each of the receiver's keys."
	self associationsDo: [:association | 
		(self isKeyValid: association) ifTrue: [aBlock value: association key]].! !

!TupleAttributeDictionary methodsFor: 'accessing' stamp: 'svp 8/23/2002 17:49'!
keysDo: aBlock 
	"Evaluate aBlock for each of the receiver's keys."
	self associationsDo: [:association | 
		(self isKeyValid: association) ifTrue: 
			[aBlock value: (Array withAll: association key)]].
! !

!ExtendedAttributeDictionary methodsFor: 'accessing' stamp: 'svp 8/13/2002 11:18'!
keysSortedSafely
	"Since we anticipate have a very large number of entries, and we
	are likely to have a very heterogenous set of keys, don't attempt to
	sort the keys"

	^self fasterKeys! !

!LargeArray class methodsFor: 'instance creation' stamp: 'svp 7/11/2002 16:34'!
new: size

	^self new: size chunkSize: self defaultChunkSize! !

!LargeArray class methodsFor: 'instance creation' stamp: 'svp 7/22/2002 13:49'!
new: size chunkSize: chunkSize

	^self new: size chunkSize: chunkSize arrayClass: Array! !

!LargeArray class methodsFor: 'instance creation' stamp: 'svp 8/13/2002 13:45'!
new: size chunkSize: chunkSize arrayClass: aClass

	| basicSize |
	(basicSize := ((size - 1) // chunkSize) + 1) = 0
		ifTrue: [basicSize := 1].
	^(self basicNew: basicSize)
		initChunkSize: chunkSize size: size arrayClass: aClass;
		yourself! !

!LargeIdentityDictionary methodsFor: 'private' stamp: 'svp 8/21/2002 15:18'!
noCheckAdd: anObject

	self shouldNotImplement! !

!LargeArray methodsFor: 'accessing' stamp: 'svp 8/12/2002 20:27'!
noCheckAt: index
	| chunkIndex |

	chunkIndex := index - 1 // chunkSize + 1.
	^(self basicAt: chunkIndex)
		at: (index - (chunkIndex - 1 * chunkSize))
! !

!LargeArray methodsFor: 'accessing' stamp: 'svp 8/12/2002 20:26'!
noCheckAt: index put: value
	| chunkIndex |

	chunkIndex := index - 1 // chunkSize + 1.
	^(self basicAt: chunkIndex)
		at: (index - (chunkIndex - 1 * chunkSize)) put: value
! !

!LargeIdentityDictionary methodsFor: 'private' stamp: 'svp 8/21/2002 17:56'!
noCheckAt: key put: value
	| index |
	array at: (index := self findElementOrNil: key) put: key.
	values at: index put: value.
	tally _ tally + 1! !

!ExtendedAttributeDictionary methodsFor: 'printing' stamp: 'svp 8/22/2002 23:57'!
printElementsOn: strm
	"Since we are expecting this to have a lot of elements, don't try and sort the
	keys like the Dictionary implementation does."

	self size < 100 ifTrue: [^super printElementsOn: strm].
	strm nextPutAll: '(a lot of items)'
! !

!LargeIdentityDictionary methodsFor: 'printing' stamp: 'svp 8/22/2002 23:50'!
printElementsOn: strm
	"Since we are expecting this to have a lot of elements, don't try and sort the
	keys like the Dictionary implementation does."

	self size < 100 ifTrue: [^super printElementsOn: strm].
	strm nextPutAll: '(a lot of items)'
! !

!LargeArray methodsFor: 'printing' stamp: 'svp 8/12/2002 11:37'!
printOn: aStream

	(#(String) includes: self arrayClass name) 
		ifTrue: [^self storeOn: aStream].
	^super printOn: aStream! !

!LargeArray methodsFor: 'private' stamp: 'svp 8/23/2002 16:06'!
pvtCheckIndex: index 

	index isInteger ifFalse: [self errorNonIntegerIndex].
	index < 1 ifTrue: [self errorSubscriptBounds: index].
	index > size ifTrue: [self errorSubscriptBounds: index].
! !

!LargeArray methodsFor: 'private' stamp: 'svp 8/2/2002 11:00'!
pvtChunkIndexOf: index

	^index - 1 // chunkSize + 1.! !

!LargeArray methodsFor: 'private' stamp: 'svp 8/2/2002 11:04'!
pvtFirstIndexOf: chunkIndex

	^((chunkIndex - 1) * chunkSize) + 1! !

!LargeArray methodsFor: 'private' stamp: 'svp 8/2/2002 11:02'!
pvtIndexOf: index inChunk: chunkIndex 

	^index - (chunkIndex - 1 * chunkSize)! !

!LargeArray methodsFor: 'private' stamp: 'svp 8/2/2002 11:06'!
pvtLastIndexOf: chunkIndex

	(chunkIndex = self basicSize) ifTrue: [^self size].
	^chunkIndex  * chunkSize! !

!ExtendedAttributeDictionary methodsFor: 'private' stamp: 'svp 8/23/2002 17:51'!
rehash
	"Smalltalk rehash."
	"Override the super implementation such that we don't include associations 
	whose key is nil (indicating that the key was reclaimed by the garbage collector)."

	| newSelf |
	newSelf _ self similarSpeciesInstance: self size.
	self associationsDo: [:each | (self isKeyValid: each) ifTrue: [newSelf noCheckAdd: each]].
	array _ newSelf array
! !

!ExtendedAttributeDictionary methodsFor: 'removing' stamp: 'svp 8/23/2002 17:52'!
removeKey: key ifAbsent: aBlock 
	"Remove key (and its associated value) from the receiver. If key is not in 
	the receiver, answer the result of evaluating aBlock. Otherwise, answer 
	the value externally named by key."

	| index assoc |
	index _ self findElementOrNil: key.
	assoc _ array at: index.
	(assoc == nil or: [self isKeyInvalid: assoc]) ifTrue: [ ^ aBlock value ].
	array at: index put: nil.
	tally _ tally - 1.
	self fixCollisionsFrom: index.
	^ assoc value! !

!LargeIdentityDictionary methodsFor: 'removing' stamp: 'svp 8/21/2002 15:13'!
removeKey: key ifAbsent: aBlock 
	"Remove key (and its associated value) from the receiver. If key is not in 
	the receiver, answer the result of evaluating aBlock. Otherwise, answer 
	the value externally named by key."

	| index k v |
	index := self findElementOrNil: key.
	k := array at: index.
	k ifNil: [ ^ aBlock value ].
	array at: index put: nil.
	v := values at: index.
	values at: index put: nil.
	tally _ tally - 1.
	self fixCollisionsFrom: index.
	^ v! !

!ExtendedAttributeDictionary methodsFor: 'private' stamp: 'svp 8/23/2002 17:43'!
scanFor: anObject startingAt: start
	"svp: This scan will detect and eliminate dead values as it searches for the key. If 
	anObject is not found (meaning a nil element was encountered before finding it), 
	this scan will return the first association with a dead key if one was encountered.  
	This allows dead slots to be reused." 

	| element firstNilKey |
	firstNilKey := nil.

	"Search from (hash mod size) to the end."
	start to: array size do:
		[:index | 
		(element _ array noCheckAt: index) == nil ifTrue: [^firstNilKey ifNil: [index]].
		(self compare: element with: anObject) ifTrue: [^index].
		(self isKeyValid: element) ifFalse: 
			[firstNilKey ifNil: [firstNilKey := index].
			element value: nil]].

	"Search from 1 to where we started."
	1 to: start-1 do:
		[:index | 
		(element _ array noCheckAt: index) == nil ifTrue: [^firstNilKey ifNil: [index]].
		(self compare: element with: anObject) ifTrue: [^index].
		(self isKeyValid: element) ifFalse: 
			[firstNilKey ifNil: [firstNilKey := index].
			element value: nil]].

	^ 0  "No match AND no empty slot"! !

!LargeIdentityDictionary methodsFor: 'private' stamp: 'svp 8/23/2002 17:32'!
scanFor: anObject startingAt: start
	"Overridden here to use #noCheckAt:...while we're at it, skip the call
	to #compare:with: to shave a little extra time off." 

	| element |

	"Search from (hash mod size) to the end."
	start to: array size do:
		[:index | ((element _ array noCheckAt: index) == nil or: 
			[element == anObject])
				ifTrue: [^ index ]].

	"Search from 1 to where we started."
	1 to: start-1 do:
		[:index | ((element _ array noCheckAt: index) == nil or: 
			[element == anObject])
				ifTrue: [^ index ]].

	^ 0  "No match AND no empty slot"! !

!LargeArray methodsFor: 'private' stamp: 'svp 8/12/2002 14:57'!
similarInstance

	^self class
		new: self size 
		chunkSize: self chunkSize 
		arrayClass: self arrayClass! !

!LargeArray methodsFor: 'private' stamp: 'svp 8/12/2002 14:57'!
similarInstance: newSize

	^self class
		new: newSize 
		chunkSize: self chunkSize 
		arrayClass: self arrayClass! !

!LargeArray methodsFor: 'private' stamp: 'svp 8/12/2002 14:57'!
similarSpeciesInstance

	^self similarInstance! !

!LargeArray methodsFor: 'private' stamp: 'svp 8/12/2002 14:58'!
similarSpeciesInstance: newSize

	^self similarInstance: newSize! !

!LargeArray methodsFor: 'accessing' stamp: 'svp 7/11/2002 15:48'!
size

	^size! !

!LargeArray methodsFor: 'private' stamp: 'svp 8/12/2002 09:07'!
speciesNew

	^self species
		new: self size 
		chunkSize: self chunkSize 
		arrayClass: self arrayClass! !

!LargeArray methodsFor: 'private' stamp: 'svp 8/12/2002 09:07'!
speciesNew: newSize

	^self species
		new: newSize 
		chunkSize: self chunkSize 
		arrayClass: self arrayClass! !

!LargeArray methodsFor: 'replacement' stamp: 'svp 8/13/2002 13:39'!
startingAt: index replace: target from: start to: stop

	"This destructively replaces elements from start to stop in the target 
	starting at index in the receiver. Answer the receiver. No range checks are 
	performed."
	
	"This implementation pushes the replacement down to the
	individual chunks.  This enables the replacement operation to be
	handled by primitives, thus getting the maximum possible performance."

	| startChunk startIndex stopChunk cStart tStart tStop |
	startChunk := self pvtChunkIndexOf: index.
	startIndex := self pvtIndexOf: index inChunk: startChunk.
	stopChunk := self pvtChunkIndexOf: (index + (stop - start)).
	cStart := nil.

	startChunk to: stopChunk do: [ :chunkIndex |
		tStart := chunkIndex = startChunk 
					ifTrue: [start] 
					ifFalse: [tStart + (chunkSize - cStart + 1)].
		tStop := chunkIndex = stopChunk
					ifTrue: [stop]
					ifFalse: [chunkIndex = startChunk 
						ifTrue: [tStart + (chunkSize - startIndex)]
						ifFalse: [tStart + chunkSize - 1]].
		cStart := chunkIndex = startChunk 
					ifTrue: [startIndex] ifFalse: [1].

		(self basicAt: chunkIndex) 
			startingAt: cStart replace: target from: tStart to: tStop].  
! !

!LargeArray methodsFor: 'printing' stamp: 'svp 8/12/2002 11:36'!
storeOn: aStream

	| x |
	(#(String) includes: self arrayClass name) ifTrue: 
		[aStream nextPut: $'.
		1 to: self size do:
			[:i |
			aStream nextPut: (x _ self at: i).
			x == $' ifTrue: [aStream nextPut: x]].
		aStream nextPutAll: ''' asLargeArrayChunkSize: '.
		aStream nextPutAll: self chunkSize asString.
		^self].
	^super storeOn: aStream! !

!LargeArray class methodsFor: 'instance creation' stamp: 'svp 8/13/2002 13:19'!
stringNew: size

	^self 
		new: size 
		chunkSize: String defaultLargeArrayChunkSize 
		arrayClass: String! !

!LargeArray class methodsFor: 'instance creation' stamp: 'svp 8/13/2002 13:19'!
stringNew: size chunkSize: chunkSize

	^self 
		new: size 
		chunkSize: chunkSize 
		arrayClass: String! !

!LargeIdentityDictionary methodsFor: 'private' stamp: 'svp 8/21/2002 15:14'!
swap: oneIndex with: otherIndex
	"May be overridden by subclasses so that fixCollisions will work"

	array swap: oneIndex with: otherIndex.
	values swap: oneIndex with: otherIndex.
! !

!LargeArray class methodsFor: 'instance creation' stamp: 'svp 8/12/2002 11:22'!
withAll: aCollection

	^(self 
		new: aCollection size 
		chunkSize: aCollection species defaultLargeArrayChunkSize 
		arrayClass: aCollection species)
			 replaceFrom: 1 to: aCollection size with: aCollection! !

!LargeArray class methodsFor: 'instance creation' stamp: 'svp 8/12/2002 11:26'!
withAll: aCollection chunkSize: chunkSize

	^(self 
		new: aCollection size 
		chunkSize: chunkSize 
		arrayClass: aCollection species)
			 replaceFrom: 1 to: aCollection size with: aCollection! !

Smalltalk at: #FilePackageManager ifPresent: [:p | p registerPackage: 'Collections-Large'].!

